unit SMNameMatch;

{ 
This module is part of the EpiData freeware project for Data Documentation and Data Entry.
All information on this. see www.epidata.dk

This Module and principle programmed by Salah Mahmud, 2002
For Delphi Pascal

The module cannot be sold or redistributed for payment nor can a patent be applied.

v1.0 - 2002
In case of errors or problems contact EpiData project at Http://www.epidata.dk

}

interface
uses sysutils;

const

alpha='ABCDEFGHIJKLMNOPQRSTUVWXYZ';
digits='0123456789';
alphadigits=alpha+digits;

function _SMSoundex(const aKey: string): string;
function _SMSoundex2(aKey: string): string;
function _SMMetaphone(const aKey: string): string;
function _CleanString(const s: string): string;
procedure _SMmetaphone2(a : pchar; lg : integer; res : pchar);
function _AlphaDigitsOnly(const s: string): string;

implementation


function _SMSoundex(const aKey: string): string;
const           {ABCDEFGHIJKLMNOPQRSTUVWXYZ}
  LetterCodes = '01230120022455012623010202';
  MaxCodeLength = 4;
var
  I: Integer;
  Ch: Char;
  LastCh: Char;
begin
  Result := '';
  LastCh := #0;
  I := 1;
  while Length(Result) <> MaxCodeLength do begin
    if I > Length(aKey) then
      Result := Result + '0'
    else begin
      Ch := UpCase(aKey[I]);
      if Ch in ['A'..'Z'] then
        if Length(Result) = 0 then
          Result := Ch
        else begin
          Ch := LetterCodes[Ord(Ch) - 64];
          if (Ch <> '0') and (Ch <> LastCh) then begin
            Result := Result + Ch;
            LastCh := Ch;
          end;
        end;
      Inc(I);
    end;
  end;
end;

function _SMSoundex2(aKey: string): string;
{ This is an "improved" Soundex algorithm from Joe Celko's "SQL for Smarties",
  1995, Morgan Kaufman Publishers, Inc, pp 85-87. }
const
  Vowels = ['A', 'E', 'I', 'O', 'U'];
var
  I: Integer;
  Ch: Char;

  function TransformPrefix(aFind, aReplace: string): Boolean;
  begin
    Result := CompareText(aFind, aKey) = 0;
    if Result then
      Move(aReplace[1], aKey[1], Length(aReplace));
  end;

  procedure Transform(aFind, aReplace: string);
  var
    I, J: Integer;
    Match: Boolean;
  begin
    for I := 2 to Length(aKey) - Length(aFind) + 1 do begin
      Match := True;
      for J := 1 to Length(aFind) do
        if aKey[I] <> aFind[J] then begin
          Match := False;
          Break;
        end;
      if Match then
        Move(aReplace[1], aKey[I], Length(aReplace));
    end;
  end;
begin

  { first pass: convert to upper, strip non letters, reduce vowels }
  I := 1;
  while I <= Length(aKey) do begin
    Ch := UpCase(aKey[I]);
    if Ch in ['A'..'Z'] then begin
      if Ch in Vowels then
        Ch := 'A';
      aKey[I] := Ch;
      Inc(I);
    end
    else
      Delete(aKey, I, 1);
  end;

  { Transform prefixes }
  if not TransformPrefix('MAC', 'MCC') then
    if not TransformPrefix('KN', 'NN') then
      if not TransformPrefix('K', 'C') then
        if not TransformPrefix('PF', 'FF') then
          if not TransformPrefix('SCH', 'SSS') then
            TransformPrefix('PH', 'FF');

  { Transform letter combinations }
  Transform('DG', 'GG');
  Transform('CAAN', 'TAAN');
  Transform('D', 'T');
  Transform('NST', 'NSS');
  Transform('AV', 'AF');
  Transform('Q', 'G');
  Transform('Z', 'S');
  Transform('M', 'N');
  Transform('KN', 'NN');
  Transform('K', 'C');

  { Replace H with A unless it is surrounded by A }
  I := 2;
  while I <= Length(aKey) do begin
    if aKey[I] = 'H' then
      if aKey[I - 1] <> 'A' then
        if (I < Length(aKey)) and (aKey[I + 1] <> 'A') then
          aKey[I] := 'H'
        else
          Inc(I);  { we know it's A so no need to check for H }
    Inc(I);
  end;

  { Replace AW with A }
  I := 2;
  while I < Length(aKey) do begin
    if (aKey[I] = 'A') and (aKey[I + 1] = 'W') then
      Delete(aKey, I + 1, 1);
    Inc(I);
  end;

  { Transforms }
  Transform('PH', 'FF');
  Transform('SCH', 'SSS');

  { Drop trailing A or S chars }
  I := Length(aKey);
  while (I > 0) and (aKey[I] in ['A', 'S']) do begin
    Delete(aKey, I, 1);
    Dec(I);
  end;

  { Transform trailing NT with TT }
  if Copy(aKey, Length(aKey) - 1, 2) = 'NT' then
    aKey := Copy(aKey, 1, Length(aKey) - 2) + 'TT';

  { Strip out all but leading A's }
  I := 2;
  while I <= Length(aKey) do
    if aKey[I] = 'A' then
      Delete(aKey, I, 1)
    else
      Inc(I);

  { Collapse repeating codes }
  I := 2;
  while I <= Length(aKey) do
    if aKey[I] = aKey[I - 1] then
      Delete(aKey, I, 1)
    else
      Inc(I);

  if Length(aKey) < 4 then
    aKey := aKey + StringofChar(' ', 4 - Length(aKey));

  Result := aKey;
end;

function _SMMetaphone(const aKey: string): string;
const
  MaxCodeLength = 6;
  VowelSet = ['A', 'E', 'I', 'O', 'U'];
  NonTransformSet = ['F', 'J', 'L', 'M', 'N', 'R'];
  EIYSet = ['E', 'I', 'Y'];
var
  Ch: Char;
  I: Integer;
  KeyBuffer: array[0..256] of Char;
  KeyBufLen: Integer;    { Number of chars in buffer }
  Key: PChar;            { Pointer to start of string }
  LastCharPos: Integer;  { Position of last char in the buffer }
begin
  Result := '';

  { Retain uppercase alpha characters in buffer; the buffer will
    always have at least one #0 placeholder before and after the
    keyword.  This avoids the need to check length bounds when
    comparing previous or next letters.  }
  FillChar(KeyBuffer, SizeOf(KeyBuffer), #0);
  Key := @KeyBuffer[1];
  KeyBufLen := 0;
  for I := 1 to Length(aKey) do begin
    Ch := UpCase(aKey[I]);
    if Ch in ['A'..'Z'] then begin
      Key[KeyBufLen] := Ch;
      Inc(KeyBufLen);
    end;
  end;
  LastCharPos := KeyBufLen - 1;

  { Transform prefixes }
  if CompareMem(Key, PChar('GN'), 2) or
     CompareMem(Key, PChar('KN'), 2) or
     CompareMem(Key, PChar('PN'), 2) or
     CompareMem(Key, PChar('AE'), 2) or
     CompareMem(Key, PChar('WH'), 2) or
     CompareMem(Key, PChar('WR'), 2) then
    Inc(Key)
  else if Key[0] = 'X' then
    Key[0] := 'S';

  for I := 0 to LastCharPos do begin

    { Skip duplicating letters except for C }
    if (Key[I - 1] = Key[I]) and (Key[I] <> 'C') then
      Continue;

    { Retain nontransform letters }
    if (Key[I] in NonTransformSet) or
       ((I = 0) and (Key[I] in VowelSet)) then begin
      Result := Result + Key[I];
      Continue;
    end;

    { Apply transforms }
    case Key[I] of
      'B': { retain unless within -MB }
           if not ((I = LastCharPos) and (Key[I - 1] = 'M')) then
             Result := Result + 'B';
      'C': { drop if in -SCI-, -SCE- or -SCY- }
           if not ((Key[I - 1] = 'S') and (Key[I + 1] in EIYSet)) then
             { map to X if in -CIA- or -CH-}
             if ((Key[I + 1] = 'I') and (Key[I + 2] = 'A')) or
                (Key[I + 1] = 'H') then
               Result := Result + 'X'
             else
               { map to S if in -CE-, -CI- or -CY- }
               if Key[I + 1] in EIYSet then
                 Result := Result + 'S'
               else  { otherwise K }
                 Result := Result + 'K';
      'D': { map to J if in -DGE-, -DGI- or -DGY- }
           if (Key[I + 1] = 'G') and (Key[I + 2] in EIYSet) then
             Result := Result + 'J'
           else  { otherwise T }
             Result := Result + 'T';
      'G': { map to J if in -GE-, -GI, -GY and not GG }
           if (Key[I + 1] in EIYSet) and (Key[I - 1] <> 'G') then
             Result := Result + 'J'
           else
                    { drop if in -GH- but not at end or before a vowel }
             if not ((Key[I + 1] = 'H') and (I <> LastCharPos - 1) and not (Key[I + 2] in VowelSet)) or
                    { drop if in -GNED }
                    ((I = LastCharPos - 3) and CompareMem(@Key[I + 1], PChar('NED'), 3)) or
                    { drop if in -GN }
                    ((I = LastCHarPos - 1) and (Key[I + 1] = 'N')) or
                    { drop if in -DGE-, -DGI- or -DGY- }
                    ((Key[I - 1] = 'D') and (Key[I + 1] in EIYSet)) then
               { otherwise K }
               Result := Result + 'K';
      'H': { retain if before a vowel and not after C, G, P, S or T }
           if (Key[I + 1] in VowelSet) and
              not (Key[I - 1] in ['C', 'G', 'P', 'S', 'T']) then
             Result := Result + 'H';
      'K': { retain unless after C }
           if Key[I - 1] <> 'C' then
             Result := Result + 'K';
      'P': { map to F if before H }
           if Key[I + 1] = 'H' then
             Result := Result + 'F'
           else  { otherwise P }
             Result := Result + 'P';
      'Q': { map to K }
           Result := Result + 'K';
      'S': { map to X if in -SH-, -SIO- or -SIA- }
           if (Key[I + 1] = 'H') or
              ((Key[I + 1] = 'I') and (Key[I + 2] in ['O', 'A'])) then
             Result := Result + 'X'
           else  { otherwise S }
             Result := Result + 'S';
      'T': { map to X if in -TIA- or -TIO- }
           if (Key[I + 1] = 'I') and (Key[I + 2] in ['O', 'A']) then
             Result := Result + 'X'
           else
             { map to 0 (zero) if before H }
             if Key[I + 1] = 'H' then
               Result := Result + '0'
             else
               { drop if in -TCH- }
               if not ((Key[I + 1] = 'C') and (Key[I + 2] = 'H')) then
                 { otherwise T }
                 Result := Result + 'T';
      'V': { map to F }
           Result := Result + 'F';
      'W',
      'Y': { retain if after a vowel }
           if Key[I - 1] in VowelSet then
             Result := Result + Key[I];
      'X': { map to KS }
           Result := Result + 'KS';
      'Z': { map to S }
           Result := Result + 'S';
    end;

    { terminate if max code length is reached }
    if Length(Result) = MaxCodeLength then
      Break;
  end;
end;


procedure _SMmetaphone2(a : pchar; lg : integer; res : pchar);
var
  b, c, d, e : string;
  inp, outp : string;
  vowels, frontv, varson, dbl : string;
  excppair, nxtltr : string;
  t, ii, jj, lng, lastchr : integer;
  curltr, prevltr, nextltr, nextltr2, nextltr3 : char;
  vowelafter, vowelbefore, frontvafter, silent, hard : boolean;
  alphachr : string;
begin
  inp:=uppercase(a);
{  showmessage(inp);}
  vowels:='AEIOU';
  frontv:='EIY';
  varson:='CSPTG';
  dbl:='.';  {Lets us allow certain letters to be doubled}
  excppair:='AGKPW';
  nxtltr:='ENNNR';
  alphachr:='ABCDEFGHIJKLMNOPQRSTUVWXYZ';

  {Remove non-alpha characters}
  outp:='';
  for t:=1 to length(inp) do
    if pos(inp[t],alphachr)>0 then outp:=outp+inp[t];

  inp:=outp;
  outp:='';

  if length(inp)=0 then begin
    strcopy(res,chr(0));
    exit;
  end;

  {Check rules at beginning of word}
  if length(inp)>1 then begin
    b:=inp[1];
    c:=inp[2];
    ii:=pos(excppair,b);
    jj:=pos(nxtltr,c);
    if (ii=jj) and (ii>0) then
      inp:=copy(inp,2,length(inp)-1);
  end;

  if inp[1]='X' then inp[1]:='S';

  if (length(inp)>2) and (copy(inp,1,2)='WH') then
    inp:='W'+copy(inp,3,length(inp)-2);

  if (length(inp)>1) and (inp[length(inp)]='S') then
    inp:=copy(inp,1,length(inp)-1);

  ii:=0;
  repeat
    ii:=ii+1;
    {Main Loop!}

    silent:=false;
    hard:=false;
    curltr:=inp[ii];
    vowelbefore:=false;
    prevltr:=' ';
    if ii>1 then begin
      prevltr:=inp[ii-1];
      if pos(prevltr,vowels)>0 then
        vowelbefore:=true;
    end;

    if ((ii=1) and (pos(curltr,vowels)>0)) then begin
      outp:=outp+curltr;
      continue;
    end;

    vowelafter:=false;
    frontvafter:=false;
    nextltr:=' ';
    if ii<length(inp) then begin
      nextltr:=inp[ii+1];
      if pos(nextltr,vowels)>0 then vowelafter:=true;
      if pos(nextltr,frontv)>0 then frontvafter:=true;
    end;

    {Skip double letters EXCEPT ones in variable double}
    if pos(curltr,dbl)=0 then
      if curltr=nextltr then
        continue;

    nextltr2:=' ';
    if (length(inp)-ii)>1 then
      nextltr2:=inp[ii+2];

    nextltr3:=' ';
    if (length(inp)-ii)>2 then
      nextltr3:=inp[ii+3];

    case curltr of

      'B' : begin
              silent:=false;
              if (ii=length(inp)) and (prevltr='M') then
                silent:=true;
              if not silent then outp:=outp+curltr;
            end;

      'C' : begin
              if not ((ii>2) and (prevltr='S') and frontvafter) then
                if ((ii>1) and (nextltr='I') and (nextltr2='A')) then
                  outp:=outp+'X'
                else
                  if frontvafter then begin
                    outp:=outp+'S';
                  end else begin
                    if ((ii>2) and (prevltr='S') and (nextltr='H')) then begin
                      outp:=outp+'K';
                    end else begin
                      if nextltr='H' then begin
                        if ((ii=1) and (pos(nextltr2,vowels)=0)) then
                          outp:=outp+'K'
                        else
                          outp:=outp+'X';
                      end else begin
                        if prevltr='C' then
                          outp:=outp+'C'
                        else
                          outp:=outp+'K';
                      end;
                    end;
                  end;
              end;

      'D' : begin
              if ((nextltr='G') and (pos(nextltr2,frontv)>0)) then
                outp:=outp+'J'
              else
                outp:=outp+'T';
            end;

      'G' : begin
              silent:=false;
              if ((ii<length(inp)) and (nextltr='H') and (pos(nextltr2,vowels)=0)) then
                silent:=true;
              if ((ii=length(inp)-4) and (nextltr='N') and (nextltr2='E') and
                (nextltr3='D')) then silent:=true
              else if ((ii=length(inp)-2) and (nextltr='N')) then silent:=true;
              if (prevltr='D') and frontvafter then silent:=true;
              if prevltr='G' then
                hard:=true;

              if not silent then begin
                if frontvafter and (not hard) then
                  outp:=outp+'J'
                else
                  outp:=outp+'K';
              end;

            end;

      'H' : begin
              silent:=false;
              if pos(prevltr,varson)>0 then silent:=true;
              if vowelbefore and (not vowelafter) then silent:=true;

              if not silent then outp:=outp+curltr;
            end;


      'F', 'J', 'L', 'M', 'N', 'R' : outp:=outp+curltr;

      'K' : if prevltr<>'C' then outp:=outp+curltr;

      'P' : if nextltr='H' then outp:=outp+'F' else outp:=outp+'P';

      'Q' : outp:=outp+'K';

      'S' : begin
              if ((ii>2) and (nextltr='I') and ((nextltr2='O') or (nextltr2='A'))) then
                outp:=outp+'X';
              if (nextltr='H') then
                outp:=outp+'X'
              else
                outp:=outp+'S';
            end;

      'T' : begin
              if ((ii>0) and (nextltr='I') and ((nextltr2='O') or (nextltr2='A'))) then
                outp:=outp+'X';
              if nextltr='H' then begin
                if ((ii>1) or (pos(nextltr2,vowels)>0)) then
                  outp:=outp+'0'
                else
                  outp:=outp+'T';
              end else
                if not ((ii<length(inp)-3) and (nextltr='C') and (nextltr2='H')) then
                  outp:=outp+'T';
            end;

      'V' : outp:=outp+'F';

      'W', 'Y' : if (ii<length(inp)-1) and vowelafter then outp:=outp+curltr;

      'X' : outp:=outp+'KS';

      'Z' : outp:=outp+'S';

    end;

  until (length(outp)>lg-1) or (ii>length(inp));
  if length(outp)> 0 then strcopy(res,@outp[1])
  else strcopy(res,'');
end;

procedure _SMNYSIIS(a : pchar; lg : integer; res : pchar);
var
  b, c, d, e : string;
  inp, outp : string;
  vowels, frontv, varson, dbl : string;
  excppair, nxtltr : string;
  t, ii, jj, lng, lastchr : integer;
  curltr, prevltr, nextltr, nextltr2, nextltr3 : char;
  vowelafter, vowelbefore, frontvafter, silent, hard : boolean;
  alphachr : string;
begin
  inp:=uppercase(a);
{  showmessage(inp);}
  vowels:='AEIOU';
  frontv:='EIY';
  varson:='CSPTG';
  dbl:='.';  {Lets us allow certain letters to be doubled}
  excppair:='AGKPW';
  nxtltr:='ENNNR';
  alphachr:='ABCDEFGHIJKLMNOPQRSTUVWXYZ';

  {Remove non-alpha characters}
  outp:='';
  for t:=1 to length(inp) do
    if pos(inp[t],alphachr)>0 then outp:=outp+inp[t];

  inp:=outp;
  outp:='';

  if length(inp)=0 then begin
    strcopy(res,chr(0));
    exit;
  end;

  {Check rules at beginning of word}
  if length(inp)>1 then begin
    b:=inp[1];
    c:=inp[2];
    ii:=pos(excppair,b);
    jj:=pos(nxtltr,c);
    if (ii=jj) and (ii>0) then
      inp:=copy(inp,2,length(inp)-1);
  end;

  if inp[1]='X' then inp[1]:='S';

  if (length(inp)>2) and (copy(inp,1,2)='WH') then
    inp:='W'+copy(inp,3,length(inp)-2);

  if (length(inp)>1) and (inp[length(inp)]='S') then
    inp:=copy(inp,1,length(inp)-1);

  ii:=0;
  repeat
    ii:=ii+1;
    {Main Loop!}

    silent:=false;
    hard:=false;
    curltr:=inp[ii];
    vowelbefore:=false;
    prevltr:=' ';
    if ii>1 then begin
      prevltr:=inp[ii-1];
      if pos(prevltr,vowels)>0 then
        vowelbefore:=true;
    end;

    if ((ii=1) and (pos(curltr,vowels)>0)) then begin
      outp:=outp+curltr;
      continue;
    end;

    vowelafter:=false;
    frontvafter:=false;
    nextltr:=' ';
    if ii<length(inp) then begin
      nextltr:=inp[ii+1];
      if pos(nextltr,vowels)>0 then vowelafter:=true;
      if pos(nextltr,frontv)>0 then frontvafter:=true;
    end;

    {Skip double letters EXCEPT ones in variable double}
    if pos(curltr,dbl)=0 then
      if curltr=nextltr then
        continue;

    nextltr2:=' ';
    if (length(inp)-ii)>1 then
      nextltr2:=inp[ii+2];

    nextltr3:=' ';
    if (length(inp)-ii)>2 then
      nextltr3:=inp[ii+3];

    case curltr of

      'B' : begin
              silent:=false;
              if (ii=length(inp)) and (prevltr='M') then
                silent:=true;
              if not silent then outp:=outp+curltr;
            end;

      'C' : begin
              if not ((ii>2) and (prevltr='S') and frontvafter) then
                if ((ii>1) and (nextltr='I') and (nextltr2='A')) then
                  outp:=outp+'X'
                else
                  if frontvafter then begin
                    outp:=outp+'S';
                  end else begin
                    if ((ii>2) and (prevltr='S') and (nextltr='H')) then begin
                      outp:=outp+'K';
                    end else begin
                      if nextltr='H' then begin
                        if ((ii=1) and (pos(nextltr2,vowels)=0)) then
                          outp:=outp+'K'
                        else
                          outp:=outp+'X';
                      end else begin
                        if prevltr='C' then
                          outp:=outp+'C'
                        else
                          outp:=outp+'K';
                      end;
                    end;
                  end;
              end;

      'D' : begin
              if ((nextltr='G') and (pos(nextltr2,frontv)>0)) then
                outp:=outp+'J'
              else
                outp:=outp+'T';
            end;

      'G' : begin
              silent:=false;
              if ((ii<length(inp)) and (nextltr='H') and (pos(nextltr2,vowels)=0)) then
                silent:=true;
              if ((ii=length(inp)-4) and (nextltr='N') and (nextltr2='E') and
                (nextltr3='D')) then silent:=true
              else if ((ii=length(inp)-2) and (nextltr='N')) then silent:=true;
              if (prevltr='D') and frontvafter then silent:=true;
              if prevltr='G' then
                hard:=true;

              if not silent then begin
                if frontvafter and (not hard) then
                  outp:=outp+'J'
                else
                  outp:=outp+'K';
              end;

            end;

      'H' : begin
              silent:=false;
              if pos(prevltr,varson)>0 then silent:=true;
              if vowelbefore and (not vowelafter) then silent:=true;

              if not silent then outp:=outp+curltr;
            end;


      'F', 'J', 'L', 'M', 'N', 'R' : outp:=outp+curltr;

      'K' : if prevltr<>'C' then outp:=outp+curltr;

      'P' : if nextltr='H' then outp:=outp+'F' else outp:=outp+'P';

      'Q' : outp:=outp+'K';

      'S' : begin
              if ((ii>2) and (nextltr='I') and ((nextltr2='O') or (nextltr2='A'))) then
                outp:=outp+'X';
              if (nextltr='H') then
                outp:=outp+'X'
              else
                outp:=outp+'S';
            end;

      'T' : begin
              if ((ii>0) and (nextltr='I') and ((nextltr2='O') or (nextltr2='A'))) then
                outp:=outp+'X';
              if nextltr='H' then begin
                if ((ii>1) or (pos(nextltr2,vowels)>0)) then
                  outp:=outp+'0'
                else
                  outp:=outp+'T';
              end else
                if not ((ii<length(inp)-3) and (nextltr='C') and (nextltr2='H')) then
                  outp:=outp+'T';
            end;

      'V' : outp:=outp+'F';

      'W', 'Y' : if (ii<length(inp)-1) and vowelafter then outp:=outp+curltr;

      'X' : outp:=outp+'KS';

      'Z' : outp:=outp+'S';

    end;

  until (length(outp)>lg-1) or (ii>length(inp));
  if length(outp)> 0 then strcopy(res,@outp[1])
  else strcopy(res,'');
end;


function _CleanString(const s: string): string;
var
  i: Integer;
begin
  Result := '';
  if Length(s) < 1 then
    Exit;
  Result := S;
  I := 1;
  While I <= Length (Result) do
      if not (Result [I]  in [#32..#126]) then
        Delete (Result, I, 1) else
        Inc (I);
end;

function _AlphaDigitsOnly(const s: string): string;
var
  i, len: Integer;
begin
  Result := '';
  len := Length(s);
  if len  < 1 then Exit;
  Result := S;
  for i:= 1 to len do
    if not(Result[I] in [#39,#48..#90,#97..#122]) then
          Result[I]:=' ';
end;



end.
